# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""
This module provides an interface to allow users to easily
build out an FFMPEG command with all the correct filters
for applying text (with a background) to the rendered media.
"""
import os
import sys
import json
from subprocess import Popen, PIPE
from PIL import ImageFont


def _is_windows():
    """
    queries if the current operating system is Windows

    :rtype: bool
    """
    return sys.platform.startswith('win') or \
        sys.platform.startswith('cygwin')


def _system_font():
    """
    attempts to determine a default system font

    :rtype: str
    """
    if _is_windows():
        font_path = os.path.join(os.environ['WINDIR'], 'Fonts')
        fonts = ('arial.ttf', 'calibri.ttf', 'times.ttf')
    elif sys.platform.startswith('darwin'):
        font_path = '/System/Library/Fonts'
        fonts = ('Menlo.ttc',)
    else:
        # assuming linux
        font_path = '/usr/share/fonts/msttcorefonts'
        fonts = ('arial.ttf', 'times.ttf', 'couri.ttf')

    system_font = None
    backup = None
    for font in fonts:
        font = os.path.join(font_path, font)
        if os.path.exists(font):
            system_font = font
            break
    else:
        if os.path.exists(font_path):
            for each in os.listdir(font_path):
                ext = os.path.splitext(each)[-1]
                if ext[1:].startswith('tt'):
                    system_font = os.path.join(font_path, each)
    return system_font or backup


# Default valuues
FONT = _system_font()
FONT_SIZE = 16
FONT_COLOR = 'white'
BG_COLOR = 'black'
BG_PADDING = 5

# FFMPEG command strings
FFMPEG = ('ffmpeg -loglevel panic -i %(input)s '
          '%(filters)s %(args)s%(output)s')
FFPROBE = ('ffprobe -v quiet -print_format json -show_format '
           '-show_streams %(source)s')
BOX = 'box=1:boxborderw=%(border)d:boxcolor=%(color)s@%(opacity).1f'
DRAWTEXT = ("drawtext=text='%(text)s':x=%(x)s:y=%(y)s:fontcolor="
            "%(color)s@%(opacity).1f:fontsize=%(size)d:fontfile='%(font)s'")
TIMECODE = ("drawtext=timecode='%(text)s':timecode_rate=%(fps).2f"
            ":x=%(x)s:y=%(y)s:fontcolor="
            "%(color)s@%(opacity).1f:fontsize=%(size)d:fontfile='%(font)s'")


# Valid aligment parameters.
TOP_CENTERED = 'top_centered'
BOTTOM_CENTERED = 'bottom_centered'
TOP_LEFT = 'top_left'
BOTTOM_LEFT = 'bottom_left'
TOP_RIGHT = 'top_right'
BOTTOM_RIGHT = 'bottom_right'


class Options(dict):
    """
    Base options class.
    """
    _params = {
        'opacity': 1,
        'x_offset': 0,
        'y_offset': 0,
        'font': FONT,
        'font_size': FONT_SIZE,
        'bg_color': BG_COLOR,
        'bg_padding': BG_PADDING,
        'font_color': FONT_COLOR
    }

    def __init__(self, **kwargs):
        super(Options, self).__init__()
        params = self._params.copy()
        params.update(kwargs)
        super(Options, self).update(**params)

    def __setitem__(self, key, value):
        if key not in self._params:
            raise KeyError("Not a valid option key '%s'" % key)
        super(Options, self).update({key: value})


class FrameNumberOptions(Options):
    """
    :key int frame_offset: offset the frame numbers
    :key float opacity: opacity value (0-1)
    :key str expression: expression that would be used instead of text
    :key bool x_offset: X position offset
                         (does not apply to centered alignments)
    :key bool y_offset: Y position offset
    :key str font: path to the font file
    :key int font_size: size to render the font in
    :key str bg_color: background color of the box
    :key int bg_padding: padding between the font and box
    :key str font_color: color to render
    """

    def __init__(self, **kwargs):
        self._params.update({
            'frame_offset': 0,
            'expression': None
        })
        super(FrameNumberOptions, self).__init__(**kwargs)


class TextOptions(Options):
    """
    :key float opacity: opacity value (0-1)
    :key str expression: expression that would be used instead of text
    :key bool x_offset: X position offset
                        (does not apply to centered alignments)
    :key bool y_offset: Y position offset
    :key str font: path to the font file
    :key int font_size: size to render the font in
    :key str bg_color: background color of the box
    :key int bg_padding: padding between the font and box
    :key str font_color: color to render
    """


class TimeCodeOptions(Options):
    """
    :key int frame_offset: offset the frame numbers
    :key float fps: frame rate to calculate the timecode by
    :key float opacity: opacity value (0-1)
    :key bool x_offset: X position offset
                         (does not apply to centered alignments)
    :key bool y_offset: Y position offset
    :key str font: path to the font file
    :key int font_size: size to render the font in
    :key str bg_color: background color of the box
    :key int bg_padding: padding between the font and box
    :key str font_color: color to render
    """

    def __init__(self, **kwargs):
        self._params.update({
            'frame_offset': 0,
            'fps': 24
        })
        super(TimeCodeOptions, self).__init__(**kwargs)


class Burnins(object):
    """
    Class that provides convenience API for building filter
    flags for the FFMPEG command.
    """

    def __init__(self, source, streams=None):
        """
        :param str source: source media file
        :param [] streams: ffprobe stream data if parsed as a pre-process
        """
        self.source = source
        self.filters = {
            'drawtext': []
        }
        self._streams = streams or _streams(self.source)

    def __repr__(self):
        return '<Overlayout - %s>' % os.path.basename(self.source)

    @property
    def start_frame(self):
        """
        :rtype: int
        """
        start_time = float(self._video_stream['start_time'])
        return round(start_time * self.frame_rate)

    @property
    def end_frame(self):
        """
        :rtype: int
        """
        end_time = float(self._video_stream['duration'])
        return round(end_time * self.frame_rate)

    @property
    def frame_rate(self):
        """
        :rtype: int
        """
        data = self._video_stream
        tokens = data['r_frame_rate'].split('/')
        return int(tokens[0]) / int(tokens[1])

    @property
    def _video_stream(self):
        video_stream = None
        for each in self._streams:
            if each.get('codec_type') == 'video':
                video_stream = each
                break
        else:
            raise RuntimeError("Failed to locate video stream "
                               "from '%s'" % self.source)
        return video_stream

    @property
    def resolution(self):
        """
        :rtype: (int, int)
        """
        data = self._video_stream
        return data['width'], data['height']

    @property
    def filter_string(self):
        """
        Generates the filter string that would be applied
        to the `-vf` argument

        :rtype: str
        """
        return ','.join(self.filters['drawtext'])

    def add_timecode(self, align, options=None):
        """
        Convenience method to create the frame number expression.

        :param enum align: alignment, must use provided enum flags
        :param dict options: recommended to use TimeCodeOptions
        """
        options = options or TimeCodeOptions()
        timecode = _frames_to_timecode(options['frame_offset'],
                                       self.frame_rate)
        options = options.copy()
        if not options.get('fps'):
            options['fps'] = self.frame_rate
        self._add_burnin(timecode.replace(':', r'\:'),
                         align,
                         options,
                         TIMECODE)

    def add_frame_numbers(self, align, options=None):
        """
        Convenience method to create the frame number expression.

        :param enum align: alignment, must use provided enum flags
        :param dict options: recommended to use FrameNumberOptions
        """
        options = options or FrameNumberOptions()
        options['expression'] = r'%%{eif\:n+%d\:d}' % options['frame_offset']
        text = str(int(self.end_frame + options['frame_offset']))
        self._add_burnin(text, align, options, DRAWTEXT)

    def add_text(self, text, align, options=None):
        """
        Adding static text to a filter.

        :param str text: text to apply to the drawtext
        :param enum align: alignment, must use provided enum flags
        :param dict options: recommended to use TextOptions
        """
        options = options or TextOptions()
        self._add_burnin(text, align, options, DRAWTEXT)

    def _add_burnin(self, text, align, options, draw):
        """
        Generic method for building the filter flags.

        :param str text: text to apply to the drawtext
        :param enum align: alignment, must use provided enum flags
        :param dict options:
        """
        resolution = self.resolution
        data = {
            'text': options.get('expression') or text,
            'color': options['font_color'],
            'size': options['font_size']
        }
        data.update(options)
        data.update(_drawtext(align, resolution, text, options))
        if 'font' in data and _is_windows():
            data['font'] = data['font'].replace(os.sep, r'\\' + os.sep)
            data['font'] = data['font'].replace(':', r'\:')
        self.filters['drawtext'].append(draw % data)

        if options.get('bg_color') is not None:
            box = BOX % {
                'border': options['bg_padding'],
                'color': options['bg_color'],
                'opacity': options['opacity']
            }
            self.filters['drawtext'][-1] += ':%s' % box

    def command(self, output=None, args=None, overwrite=False):
        """
        Generate the entire FFMPEG command.

        :param str output: output file
        :param str args: additional FFMPEG arguments
        :param bool overwrite: overwrite the output if it exists
        :returns: completed command
        :rtype: str
        """
        output = output or ''
        if overwrite:
            output = '-y %s' % output
        return (FFMPEG % {
            'input': self.source,
            'output': output,
            'args': '%s ' % args if args else '',
            'filters': '-vf "%s"' % self.filter_string
        }).strip()

    def render(self, output, args=None, overwrite=False):
        """
        Render the media to a specified destination.

        :param str output: output file
        :param str args: additional FFMPEG arguments
        :param bool overwrite: overwrite the output if it exists
        """
        if not overwrite and os.path.exists(output):
            raise RuntimeError("Destination '%s' exists, please "
                               "use overwrite" % output)
        command = self.command(output=output,
                               args=args,
                               overwrite=overwrite)
        proc = Popen(command, shell=True)
        proc.communicate()
        if proc.returncode != 0:
            raise RuntimeError("Failed to render '%s': %s'"
                               % (output, command))
        if not os.path.exists(output):
            raise RuntimeError("Failed to generate '%s'" % output)


def _streams(source):
    """
    :param str source: source media file
    :rtype: [{}, ...]
    """
    command = FFPROBE % {'source': source}
    proc = Popen(command, shell=True, stdout=PIPE)
    out = proc.communicate()[0]
    if proc.returncode != 0:
        raise RuntimeError("Failed to run: %s" % command)
    return json.loads(out)['streams']


def _drawtext(align, resolution, text, options):
    """
    :rtype: {'x': int, 'y': int}
    """
    x_pos = '0'
    if align in (TOP_CENTERED, BOTTOM_CENTERED):
        x_pos = 'w/2-tw/2'
    elif align in (TOP_RIGHT, BOTTOM_RIGHT):
        ifont = ImageFont.truetype(options['font'],
                                   options['font_size'])
        box_size = ifont.getsize(text)
        x_pos = resolution[0] - (box_size[0] + options['x_offset'])
    elif align in (TOP_LEFT, BOTTOM_LEFT):
        x_pos = options['x_offset']

    if align in (TOP_CENTERED,
                 TOP_RIGHT,
                 TOP_LEFT):
        y_pos = '%d' % options['y_offset']
    else:
        y_pos = 'h-text_h-%d' % (options['y_offset'])
    return {'x': x_pos, 'y': y_pos}


def _frames_to_timecode(frames, framerate):
    return '{0:02d}:{1:02d}:{2:02d}:{3:02d}'.format(
        int(frames / (3600 * framerate)),
        int(frames / (60 * framerate) % 60),
        int(frames / framerate % 60),
        int(frames % framerate))



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""FFMPEG Burnins Adapter"""
import os
import sys


def build_burnins(input_otio):
    """
    Generates the burnin objects for each clip within the otio container

    :param input_otio: OTIO container
    :rtype: [ffmpeg_burnins.Burnins(), ...]
    """

    if os.path.dirname(__file__) not in sys.path:
        sys.path.append(os.path.dirname(__file__))

    import ffmpeg_burnins
    key = 'burnins'

    burnins = []
    for clip in input_otio.each_clip():

        # per clip burnin data
        burnin_data = clip.media_reference.metadata.get(key)
        if not burnin_data:
            # otherwise default to global burnin
            burnin_data = input_otio.metadata.get(key)

        if not burnin_data:
            continue

        media = clip.media_reference.target_url
        if media.startswith('file://'):
            media = media[7:]
        streams = burnin_data.get('streams')
        burnins.append(ffmpeg_burnins.Burnins(media,
                                              streams=streams))
        burnins[-1].otio_media = media
        burnins[-1].otio_overwrite = burnin_data.get('overwrite')
        burnins[-1].otio_args = burnin_data.get('args')

        for burnin in burnin_data.get('burnins', []):
            align = burnin.pop('align')
            function = burnin.pop('function')
            if function == 'text':
                text = burnin.pop('text')
                options = ffmpeg_burnins.TextOptions()
                options.update(burnin)
                burnins[-1].add_text(text, align, options=options)
            elif function == 'frame_number':
                options = ffmpeg_burnins.FrameNumberOptions()
                options.update(burnin)
                burnins[-1].add_frame_numbers(align, options=options)
            elif function == 'timecode':
                options = ffmpeg_burnins.TimeCodeOptions()
                options.update(burnin)
                burnins[-1].add_timecode(align, options=options)
            else:
                raise RuntimeError("Unknown function '%s'" % function)

    return burnins


def write_to_file(input_otio, filepath):
    """required OTIO function hook"""

    for burnin in build_burnins(input_otio):
        burnin.render(os.path.join(filepath, burnin.otio_media),
                      args=burnin.otio_args,
                      overwrite=burnin.otio_overwrite)



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""RV External Adapter component.

Because the rv adapter requires being run from within the RV py-interp to take
advantage of modules inside of RV, this script gets shelled out to from the
RV OTIO adapter.

Requires that you set the environment variables:
    OTIO_RV_PYTHON_LIB - should point at the parent directory of rvSession
    OTIO_RV_PYTHON_BIN - should point at py-interp from within rv
"""

# python
import sys
import os
import json

# rv import
sys.path += [os.path.join(os.environ["OTIO_RV_PYTHON_LIB"], "rvSession")]
import rvSession  # noqa


_RV_TYPE_MAP = {
    "rvSession.gto.FLOAT": rvSession.gto.FLOAT,
    "rvSession.gto.STRING": rvSession.gto.STRING,
}


# because json.loads returns a unicode type
_UNICODE_TYPE = type(u"")


def main():
    """ entry point, should be called from the rv adapter in otio """

    session_file = rvSession.Session()

    output_fname = sys.argv[1]

    simplified_data = json.loads(sys.stdin.read())
    if sys.version_info.major <= 2:
        simplified_data = _remove_unicode(simplified_data)

    result = execute_rv_commands(simplified_data, session_file)

    session_file.setViewNode(result)
    session_file.write(output_fname)


def execute_rv_commands(simplified_data, to_session):
    rv_nodes = []
    for node in simplified_data["nodes"]:
        new_node = to_session.newNode(str(node["kind"]), str(node["name"]))
        rv_node_index = len(rv_nodes)

        # make sure that node order lines up
        assert rv_node_index == node["node_index"]

        rv_nodes.append(new_node)
        node["rv_node"] = new_node

        for prop in node["properties"]:
            args = prop
            # the fourth argument is the type
            args[4] = _RV_TYPE_MAP[args[4]]

            new_node.setProperty(*args)

        for (fn, args) in node["commands"]:
            getattr(new_node, fn)(args)

    # inputs done as a second pass now that all nodes are created
    for node in simplified_data["nodes"]:
        for input in node["inputs"]:
            node["rv_node"].addInput(rv_nodes[input])

    # return the first node created.
    return rv_nodes[0]


def _remove_unicode(blob):
    if _UNICODE_TYPE == type(blob):
        return blob.encode('utf-8')

    if isinstance(blob, dict):
        result = {}
        for key, val in blob.items():
            result[_remove_unicode(key)] = _remove_unicode(val)
        return result

    if isinstance(blob, list):
        return [_remove_unicode(i) for i in blob]

    return blob


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

try:
    from PySide6 import QtGui, QtCore, QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets
from collections import OrderedDict, namedtuple

import opentimelineio as otio
from . import (
    ruler_widget,
    track_widgets,
)


KEY_SYM = {
    QtCore.Qt.Key_Left: QtCore.Qt.Key_Right,
    QtCore.Qt.Key_Right: QtCore.Qt.Key_Left,
    QtCore.Qt.Key_Up: QtCore.Qt.Key_Down,
    QtCore.Qt.Key_Down: QtCore.Qt.Key_Up
}


def get_nav_menu_data():
    _nav_menu = namedtuple(
        "nav_menu",
        ["bitmask", "otioItem", "default", "exclusive"]
    )

    filter_dict = OrderedDict(
        [
            (
                "Clip",
                _nav_menu(0b00000001, track_widgets.ClipItem, True, False)
            ),
            (
                "Nested Clip",
                _nav_menu(0b00000010, track_widgets.NestedItem, True, False)
            ),
            (
                "Gap",
                _nav_menu(0b00000100, track_widgets.GapItem, True, False)
            ),
            (
                "Transition",
                _nav_menu(0b00001000, track_widgets.TransitionItem, True, False)
            ),
            (
                "Only with Marker",
                _nav_menu(0b00010000, track_widgets.Marker, False, True)
            ),
            (
                "Only with Effect",
                _nav_menu(0b00100000, track_widgets.EffectItem, False, True)
            ),
            # ("All", nav_menu(0b01000000, "None", False)) @TODO
        ]
    )
    return filter_dict


def get_filters(filter_dict, bitmask):
    filters = list()
    for item in filter_dict.values():
        if bitmask & item.bitmask:
            filters.append(item)
    return filters


def build_menu(navigation_menu):
    filter_dict = get_nav_menu_data()
    actions = list()
    for label, nav in filter_dict.items():
        if label == "Only with Marker":
            navigation_menu.addSeparator()

        action = navigation_menu.addAction(label)
        action.setCheckable(True)
        action.setChecked(nav.default)
        actions.append(action)

    return actions


def group_filters(bitmask):
    inclusive_filters = list()
    exclusive_filters = list()
    filter_dict = get_nav_menu_data()
    for item in get_filters(filter_dict, bitmask):
        if item.exclusive:
            exclusive_filters.append(item)
        else:
            inclusive_filters.append(item)
    return inclusive_filters, exclusive_filters


class CompositionWidget(QtWidgets.QGraphicsScene):

    def __init__(self, composition, *args, **kwargs):
        super(CompositionWidget, self).__init__(*args, **kwargs)
        self.composition = composition
        self.setBackgroundBrush(QtGui.QBrush(QtGui.QColor(33, 33, 33)))

        self._adjust_scene_size()
        self._add_tracks()
        self._add_time_slider()
        self._add_markers()
        self._ruler = self._add_ruler()

        self._data_cache = self._cache_tracks()

    def track_name_width(self):
        # Choose track name item from the longest track, since that track will
        # define the composition's scene width.
        width = track_widgets.TRACK_NAME_WIDGET_WIDTH
        longest_track = None
        max_end_time = 0

        for item in self.items():
            if isinstance(item, track_widgets.Track):
                track_range = item.track.trimmed_range_in_parent()
                end_time = track_range.end_time_inclusive().to_seconds()
                if end_time > max_end_time:
                    max_end_time = end_time
                    longest_track = item

        if longest_track is not None:
            if longest_track.track_name_item is not None:
                width = longest_track.track_name_item.rect().width()

        return width

    def _adjust_scene_size(self, zoom_level=1.0):
        scene_range = self.composition.trimmed_range()

        start_time = otio.opentime.to_seconds(scene_range.start_time)
        duration = otio.opentime.to_seconds(scene_range.end_time_exclusive())

        if isinstance(self.composition, otio.schema.Stack):
            # non audio tracks are sorted into one area
            has_video_tracks = any(
                t.kind != otio.schema.TrackKind.Audio
                for t in self.composition
            )
            has_audio_tracks = any(
                t.kind == otio.schema.TrackKind.Audio
                for t in self.composition
            )
        elif isinstance(self.composition, otio.schema.Track):
            has_video_tracks = (
                self.composition.kind != otio.schema.TrackKind.Audio
            )
            has_audio_tracks = (
                self.composition.kind == otio.schema.TrackKind.Audio
            )
        else:
            raise otio.exceptions.NotSupportedError(
                "Error: file includes composition '{}', of type '{}',"
                " not supported by opentimeview.  Only supports children of"
                " otio.schema.Stack and otio.schema.Track".format(
                    self.composition,
                    type(self.composition)
                )
            )

        height = (
            track_widgets.TIME_SLIDER_HEIGHT +
            (
                int(has_video_tracks and has_audio_tracks) *
                track_widgets.MEDIA_TYPE_SEPARATOR_HEIGHT
            ) +
            len(self.composition) * track_widgets.TRACK_HEIGHT
        )

        self.setSceneRect(
            start_time * track_widgets.TIME_MULTIPLIER,
            0,
            duration * track_widgets.TIME_MULTIPLIER +
            self.track_name_width() * zoom_level,
            height
        )

    def counteract_zoom(self, zoom_level=1.0):
        # some items we do want to keep the same visual size. So we need to
        # inverse the effect of the zoom
        items_to_scale = [
            i for i in self.items()
            if (isinstance(i, (track_widgets.BaseItem, track_widgets.Marker,
                               ruler_widget.Ruler, track_widgets.TimeSlider)))
        ]

        for item in items_to_scale:
            item.counteract_zoom(zoom_level)

        self._adjust_scene_size(zoom_level)

    def _add_time_slider(self):
        scene_rect = self.sceneRect()
        scene_rect.setWidth(scene_rect.width() * 10)
        scene_rect.setHeight(track_widgets.TIME_SLIDER_HEIGHT)
        self._time_slider = track_widgets.TimeSlider(scene_rect)
        self.addItem(self._time_slider)
        # Make sure that the ruler is on top of the selected Track items
        self._time_slider.setZValue(float('inf'))

    def _add_track(self, track, y_pos):
        scene_rect = self.sceneRect()
        rect = QtCore.QRectF(0, 0, scene_rect.width() * 10,
                             track_widgets.TRACK_HEIGHT)
        new_track = track_widgets.Track(track, rect)
        self.addItem(new_track)
        new_track.setPos(scene_rect.x(), y_pos)

    def _add_tracks(self):
        video_tracks_top = track_widgets.TIME_SLIDER_HEIGHT
        audio_tracks_top = track_widgets.TIME_SLIDER_HEIGHT

        video_tracks = []
        audio_tracks = []
        other_tracks = []

        if isinstance(self.composition, otio.schema.Stack):
            video_tracks = [
                t for t in self.composition
                if t.kind == otio.schema.TrackKind.Video
            ]
            audio_tracks = [
                t for t in self.composition
                if t.kind == otio.schema.TrackKind.Audio
            ]
            video_tracks.reverse()

            other_tracks = [
                t for t in self.composition
                if (
                    t.kind not in (
                        otio.schema.TrackKind.Video,
                        otio.schema.TrackKind.Audio
                    )
                )
            ]
        else:
            if self.composition.kind == otio.schema.TrackKind.Video:
                video_tracks = [self.composition]
            elif self.composition.kind == otio.schema.TrackKind.Audio:
                audio_tracks = [self.composition]
            else:
                other_tracks = [self.composition]

        if other_tracks:
            for t in other_tracks:
                print(
                    "Warning: track named '{}' has nonstandard track type:"
                    " '{}'".format(t.name, t.kind)
                )

            video_tracks.extend(other_tracks)

        video_tracks_top = track_widgets.TIME_SLIDER_HEIGHT
        audio_tracks_top = (
            track_widgets.TIME_SLIDER_HEIGHT +
            len(video_tracks) * track_widgets.TRACK_HEIGHT +
            int(
                bool(video_tracks) and bool(audio_tracks)
            ) * track_widgets.MEDIA_TYPE_SEPARATOR_HEIGHT
        )

        for i, track in enumerate(audio_tracks):
            self._add_track(track, audio_tracks_top + i *
                            track_widgets.TRACK_HEIGHT)

        for i, track in enumerate(video_tracks):
            self._add_track(track, video_tracks_top + i *
                            track_widgets.TRACK_HEIGHT)

    def _add_markers(self):
        for m in self.composition.markers:
            marker = track_widgets.Marker(m, None)
            marker.setX(
                otio.opentime.to_seconds(m.marked_range.start_time) *
                track_widgets.TIME_MULTIPLIER
            )
            marker.setY(track_widgets.TIME_SLIDER_HEIGHT -
                        track_widgets.MARKER_SIZE)
            marker.setParentItem(self._time_slider)
            marker.counteract_zoom()

    def _add_ruler(self):
        scene_rect = self.sceneRect()
        ruler = ruler_widget.Ruler(scene_rect.height(), composition=self)
        ruler.setParentItem(self._time_slider)
        ruler.setX(scene_rect.width() / 2)
        ruler.setY(track_widgets.TIME_SLIDER_HEIGHT -
                   track_widgets.MARKER_SIZE)

        ruler.counteract_zoom()
        return ruler

    def get_ruler(self):
        return self._ruler

    def get_next_item(self, item, key):
        otio_item = item.item
        next_item = None
        if key in [QtCore.Qt.Key_Right, QtCore.Qt.Key_Left]:
            head, tail = otio_item.parent().neighbors_of(otio_item)
            next_item = head if key == QtCore.Qt.Key_Left else tail
        elif key in [QtCore.Qt.Key_Up, QtCore.Qt.Key_Down]:
            track = item.parentItem()
            if self._data_cache[track][key]:
                next_track = self._data_cache[track][key]
            else:
                # No more track in that direction.
                return item
            at_time = otio_item.trimmed_range_in_parent().start_time

            # If at_time is out of track_range, then get the latest item
            # in the track.
            at_time = min(at_time,
                          self._data_cache[next_track]["end_time_inclusive"])

            next_item = next_track.track.child_at_time(at_time)

        # need the widget, not the otio instance.
        if next_item:
            next_item = self._data_cache["map_to_widget"][next_item]

        return next_item

    def _get_closest_item(self, item, start_time, filters):
        next_item = item
        tail = self.get_next_item_filters(item, QtCore.Qt.Key_Right, filters)
        head = self.get_next_item_filters(item, QtCore.Qt.Key_Left, filters)

        # 3 cases
        # 1. tail and head are both valid => get the closest one
        if not tail == item and not head == item:
            tail_start_time = tail.item.trimmed_range_in_parent().start_time
            head_end_time = head.item.trimmed_range_in_parent().\
                end_time_inclusive()

            time_to_tail = tail_start_time - start_time
            time_to_head = start_time - head_end_time

            next_item = tail if time_to_tail < time_to_head else head

        # 2. only tail is valid
        elif not tail == item and head == item:
            next_item = tail

        # 3. only head is valid
        elif tail == item and not head == item:
            next_item = head

        return next_item

    def get_next_item_filters(self, item, key, filters):
        original_item = item
        next_item = None
        while not match_filters(next_item, filters):
            next_item = self.get_next_item(item, key)

            if key in [QtCore.Qt.Key_Up, QtCore.Qt.Key_Down] and \
               not match_filters(next_item, filters):
                start_time = item.item.trimmed_range_in_parent().start_time
                next_item = self._get_closest_item(next_item, start_time,
                                                   filters)

            if next_item is None or next_item == item:
                next_item = original_item
                break

            item = next_item

        return next_item

    def _cache_tracks(self):
        '''
        Create a doubly linked list to navigate from track to track:
            track->get_next_up & track->get_next_up
        "map_to_widget" : Create a map to retrieve the pyside widget from
        the otio item
        '''
        data_cache = dict()
        tracks = list()
        data_cache["map_to_widget"] = dict()
        for track_item in self.items():
            if not isinstance(track_item, track_widgets.Track):
                continue
            tracks.append(track_item)
            track_range = track_item.track.available_range()
            data_cache[track_item] = {QtCore.Qt.Key_Up: None,
                                      QtCore.Qt.Key_Down: None,
                                      "end_time_inclusive":
                                      track_range.end_time_inclusive()
                                      }

            for item in track_item.childItems():
                data_cache["map_to_widget"][item.item] = item

        tracks.sort(key=lambda y: y.pos().y())
        index_last_track = len(tracks) - 1
        for i, track_item in enumerate(tracks):
            data_cache[track_item][QtCore.Qt.Key_Up] = \
                tracks[i - 1] if i > 0 else None
            data_cache[track_item][QtCore.Qt.Key_Down] = \
                tracks[i + 1] if i < index_last_track else None

        return data_cache


def match_filters(item, filters=None):
    if not item:
        return None
    if filters is None:
        return item

    otio_children = list()
    if isinstance(item, track_widgets.BaseItem):
        otio_children = item.get_otio_sub_items()

    for filter in filters.exclusive:
        excl_item = [i for i in otio_children
                     if isinstance(i, filter.otioItem)]
        if not excl_item:
            return None

    for filter in filters.inclusive:
        if isinstance(item, filter.otioItem):
            return item

    return None


class CompositionView(QtWidgets.QGraphicsView):

    open_stack = QtCore.Signal(otio.schema.Stack)
    selection_changed = QtCore.Signal(otio.core.SerializableObject)

    def __init__(self, stack, *args, **kwargs):
        super(CompositionView, self).__init__(*args, **kwargs)
        self.setResizeAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.setScene(CompositionWidget(stack, parent=self))
        self.setAlignment((QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop))
        self.setStyleSheet('border: 0px;')
        self.scene().selectionChanged.connect(self.parse_selection_change)
        self._navigation_filter = None
        self._last_item_cache = {"key": None, "item": None,
                                 "previous_item": None}

    def parse_selection_change(self):
        selection = self.scene().selectedItems()
        if not selection:
            return
        # Exclude ruler from selection
        for item in selection:
            if isinstance(item, ruler_widget.Ruler):
                continue
            self.selection_changed.emit(item.item)
            break

    def mousePressEvent(self, mouse_event):
        modifiers = QtWidgets.QApplication.keyboardModifiers()
        self.setDragMode(
            QtWidgets.QGraphicsView.ScrollHandDrag
            if modifiers == QtCore.Qt.AltModifier
            else QtWidgets.QGraphicsView.NoDrag
        )
        self.setInteractive(not modifiers == QtCore.Qt.AltModifier)

        super(CompositionView, self).mousePressEvent(mouse_event)

    def mouseReleaseEvent(self, mouse_event):
        super(CompositionView, self).mouseReleaseEvent(mouse_event)
        self.setDragMode(QtWidgets.QGraphicsView.NoDrag)

    def wheelEvent(self, event):
        try:
            # PySide6:
            # https://doc.qt.io/qtforpython/PySide6/QtGui/QWheelEvent.html
            delta_point = event.angleDelta()
            delta = delta_point.y()

        except AttributeError:
            # PySide2:
            # https://doc.qt.io/qtforpython-5/PySide2/QtGui/QWheelEvent.html
            delta = event.delta()

        scale_by = 1.0 + float(delta) / 1000
        self.scale(scale_by, 1)
        zoom_level = 1.0 / self.transform().m11()
        track_widgets.CURRENT_ZOOM_LEVEL = zoom_level
        self.scene().counteract_zoom(zoom_level)

    def _get_first_item(self):
        newXpos = 0
        newYpos = track_widgets.TIME_SLIDER_HEIGHT

        newPosition = QtCore.QPointF(newXpos, newYpos)

        return self.scene().itemAt(newPosition, QtGui.QTransform())

    def _get_left_item(self, curSelectedItem):
        curItemXpos = curSelectedItem.pos().x()

        if curSelectedItem.parentItem():
            curTrackYpos = curSelectedItem.parentItem().pos().y()

            newXpos = curItemXpos - 1
            newYpos = curTrackYpos

            if newXpos < 0:
                newXpos = 0
        else:
            newXpos = curItemXpos
            newYpos = curSelectedItem.y()

        newPosition = QtCore.QPointF(newXpos, newYpos)

        return self.scene().itemAt(newPosition, QtGui.QTransform())

    def _get_right_item(self, curSelectedItem):
        curItemXpos = curSelectedItem.pos().x()

        if curSelectedItem.parentItem():
            curTrackYpos = curSelectedItem.parentItem().pos().y()

            newXpos = curItemXpos + curSelectedItem.rect().width()
            newYpos = curTrackYpos
        else:
            newXpos = curItemXpos
            newYpos = curSelectedItem.y()

        newPosition = QtCore.QPointF(newXpos, newYpos)

        return self.scene().itemAt(newPosition, QtGui.QTransform())

    def _get_up_item(self, curSelectedItem):
        curItemXpos = curSelectedItem.pos().x()

        if curSelectedItem.parentItem():
            curTrackYpos = curSelectedItem.parentItem().pos().y()

            newXpos = curItemXpos
            newYpos = curTrackYpos - track_widgets.TRACK_HEIGHT

            newSelectedItem = self.scene().itemAt(
                QtCore.QPointF(
                    newXpos,
                    newYpos
                ),
                QtGui.QTransform()
            )

            if (
                    not newSelectedItem
                    or isinstance(newSelectedItem, otio.schema.Track)
            ):
                newYpos = newYpos - track_widgets.TRANSITION_HEIGHT
        else:
            newXpos = curItemXpos
            newYpos = curSelectedItem.y()

        newPosition = QtCore.QPointF(newXpos, newYpos)

        return self.scene().itemAt(newPosition, QtGui.QTransform())

    def _get_down_item(self, curSelectedItem):
        curItemXpos = curSelectedItem.pos().x()

        if curSelectedItem.parentItem():
            curTrackYpos = curSelectedItem.parentItem().pos().y()
            newXpos = curItemXpos
            newYpos = curTrackYpos + track_widgets.TRACK_HEIGHT

            newSelectedItem = self.scene().itemAt(
                QtCore.QPointF(
                    newXpos,
                    newYpos
                ),
                QtGui.QTransform()
            )

            if (
                    not newSelectedItem
                    or isinstance(newSelectedItem, otio.schema.Track)
            ):
                newYpos = newYpos + track_widgets.TRANSITION_HEIGHT

            if newYpos < track_widgets.TRACK_HEIGHT:
                newYpos = track_widgets.TRACK_HEIGHT
        else:
            newXpos = curItemXpos
            newYpos = (
                track_widgets.MARKER_SIZE + track_widgets.TIME_SLIDER_HEIGHT + 1
            )
            newYpos = track_widgets.TIME_SLIDER_HEIGHT
        newPosition = QtCore.QPointF(newXpos, newYpos)

        return self.scene().itemAt(newPosition, QtGui.QTransform())

    def _deselect_all_items(self):
        if self.scene().selectedItems:
            for selectedItem in self.scene().selectedItems():
                selectedItem.setSelected(False)

    def _select_new_item(self, newSelectedItem):
        # Check for text item
        # Text item shouldn't be selected,
        # maybe a bug in the population of timeline.
        if isinstance(newSelectedItem, QtWidgets.QGraphicsSimpleTextItem):
            newSelectedItem = newSelectedItem.parentItem()

        # Validate new item for edge cases
        # If valid, set selected
        if (
            not isinstance(newSelectedItem, track_widgets.Track) and
            newSelectedItem
        ):
            self._deselect_all_items()
            newSelectedItem.setSelected(True)
            self.centerOn(newSelectedItem)

    def _get_new_item(self, key_event, curSelectedItem):
        key = key_event.key()
        modifier = key_event.modifiers()
        if not (key in (
                QtCore.Qt.Key_Left,
                QtCore.Qt.Key_Right,
                QtCore.Qt.Key_Up,
                QtCore.Qt.Key_Down,
                QtCore.Qt.Key_Return,
                QtCore.Qt.Key_Enter
        ) and not (modifier & QtCore.Qt.ControlModifier)):
            return None

        if key in [QtCore.Qt.Key_Left, QtCore.Qt.Key_Right,
                   QtCore.Qt.Key_Up, QtCore.Qt.Key_Down]:

            # check if last key was the opposite direction of current one
            if KEY_SYM[key] == self._last_item_cache["key"] and \
               curSelectedItem == self._last_item_cache["item"] and \
               not curSelectedItem == self._last_item_cache["previous_item"]:
                newSelectedItem = self._last_item_cache["previous_item"]
            else:
                filters = self.get_filters()
                # make sure that the selected item is a BaseItem
                while (not
                       isinstance(curSelectedItem, track_widgets.BaseItem) and
                       curSelectedItem):
                    curSelectedItem = curSelectedItem.parentItem()
                if not curSelectedItem:
                    return None

                newSelectedItem = self.scene().get_next_item_filters(
                    curSelectedItem,
                    key,
                    filters
                )

            # self._last_item_cache["item"] = curSelectedItem
            self._last_item_cache["item"] = newSelectedItem
            self._last_item_cache["previous_item"] = curSelectedItem
            self._last_item_cache["key"] = key
        elif key in [QtCore.Qt.Key_Return, QtCore.Qt.Key_Return]:
            if isinstance(curSelectedItem, track_widgets.NestedItem):
                curSelectedItem.keyPressEvent(key_event)
                newSelectedItem = None

        return newSelectedItem

    def keyPressEvent(self, key_event):
        super(CompositionView, self).keyPressEvent(key_event)
        self.setInteractive(True)

        # Remove ruler_widget.Ruler instance from selection
        selections = [
            x for x in self.scene().selectedItems()
            if not isinstance(x, ruler_widget.Ruler)
        ]

        # Based on direction key, select new selected item
        if not selections:
            newSelectedItem = self._get_first_item()
        # No item selected, so select the first item
        else:
            curSelectedItem = selections[0]

            # Check to see if the current selected item is a rect item
            # If current selected item is not a rect, then extra tests
            # are needed.
            if not isinstance(curSelectedItem, QtWidgets.QGraphicsRectItem):
                if curSelectedItem.parentItem():
                    curSelectedItem = curSelectedItem.parentItem()

            newSelectedItem = self._get_new_item(key_event, curSelectedItem)
            self._keyPress_frame_all(key_event)
            self._snap(key_event, curSelectedItem)
        if newSelectedItem:
            self._select_new_item(newSelectedItem)

    def _snap(self, key_event, curSelectedItem):
        key = key_event.key()
        modifier = key_event.modifiers()
        if key in (
                QtCore.Qt.Key_Left,
                QtCore.Qt.Key_Right,
        ) and (modifier & QtCore.Qt.ControlModifier):
            direction = 0
            if key == QtCore.Qt.Key_Left:
                direction = -1.0
            elif key == QtCore.Qt.Key_Right:
                direction = 1.0
            if direction:
                ruler = self.scene().get_ruler()
                ruler.snap(direction=direction,
                           scene_width=self.sceneRect().width())
                self.ensureVisible(ruler)

    def _keyPress_frame_all(self, key_event):
        key = key_event.key()
        modifier = key_event.modifiers()
        if key == QtCore.Qt.Key_F and (modifier & QtCore.Qt.ControlModifier):
            self.frame_all()

    def frame_all(self):
        self.resetTransform()
        track_widgets.CURRENT_ZOOM_LEVEL = 1.0
        self.scene().counteract_zoom()

        track_name_width = self.scene().track_name_width()
        view_width = self.size().width() - track_name_width
        scene_width = self.sceneRect().width() - track_name_width
        if not view_width or not scene_width:
            # Prevent zero division errors
            return

        scale_by = view_width / scene_width
        self.scale(scale_by, 1)
        zoom_level = 1.0 / self.transform().m11()
        track_widgets.CURRENT_ZOOM_LEVEL = zoom_level
        self.scene().counteract_zoom(zoom_level)

    def navigationfilter_changed(self, bitmask):
        '''
        Update the navigation filter according to the filters checked in the
        navigation menu. Reset _last_item_cache
        '''
        nav_d = namedtuple("navigation_filter", ["inclusive", "exclusive"])
        incl_filter, excl_filter = group_filters(bitmask)
        self._navigation_filter = nav_d(incl_filter, excl_filter)
        self._last_item_cache = {"key": None, "item": None,
                                 "previous_item": None}

    def get_filters(self):
        return self._navigation_filter


class Timeline(QtWidgets.QTabWidget):

    selection_changed = QtCore.Signal(otio.core.SerializableObject)
    navigationfilter_changed = QtCore.Signal(int)

    def __init__(self, *args, **kwargs):
        super(Timeline, self).__init__(*args, **kwargs)
        self.timeline = None

        self.setTabsClosable(True)
        self.tabCloseRequested.connect(self._close_tab)

    def _close_tab(self, index):
        self.widget(index).deleteLater()
        self.removeTab(index)

    def set_timeline(self, timeline):
        # close all the tabs
        for i in reversed(range(self.count())):
            self._close_tab(i)

        # load new timeline
        self.timeline = timeline
        if timeline is not None:
            self.add_stack(timeline.tracks)

    def add_stack(self, stack):
        """Open a tab for the stack or go to it if already present"""

        # find the tab for the stack if the tab has already been opened
        tab_index = next(
            (
                i for i in range(self.count())
                if stack is self.widget(i).scene().composition
            ),
            None
        )

        if tab_index is not None:
            self.setCurrentIndex(tab_index)
            return

        new_stack = CompositionView(stack, parent=self)
        self.addTab(new_stack, stack.name)

        # cannot close the first tab
        if self.count() == 1:
            button = self.tabBar().tabButton(0, QtWidgets.QTabBar.RightSide)
            if button:
                button.resize(0, 0)

        new_stack.open_stack.connect(self.add_stack)
        new_stack.selection_changed.connect(self.selection_changed)
        self.navigationfilter_changed.connect(
            new_stack.navigationfilter_changed
        )
        self.setCurrentIndex(self.count() - 1)
        self.frame_all()

    def frame_all(self):
        if self.currentWidget():
            self.currentWidget().frame_all()



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

# flake8: noqa

from . import (
    details_widget,
    timeline_widget,
)

__version__ = "0.15.0"
__author__ = "Contributors to the OpenTimelineIO project"
__author_email__ = "otio-discussion@lists.aswf.io"
__license__ = "Modified Apache 2.0 License"



# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""OTIO to SVG Adapter
   Points in calculations are y-up.
   Points in SVG are y-down."""

# otio
import opentimelineio as otio
from xml.etree.ElementTree import Element, SubElement, tostring
from xml.dom import minidom

# python
import math
import random

random_colors_used = []


class Color:
    def __init__(self, r=0.0, g=0.0, b=0.0, a=255.0):
        self.value = (r, g, b, a)

    def __getitem__(self, item):
        return self.value[item]

    @staticmethod
    def random_color():
        color = Color.__generate_new_color()
        random_colors_used.append(color)
        return color

    @staticmethod
    def __generate_new_color():
        max_distance = None
        best_color = None
        for _ in range(100):
            color = Color.__get_random_color()
            if len(random_colors_used) == 0:
                return color
            best_distance = min([Color.__color_distance(color, c)
                                 for c in random_colors_used])
            if not max_distance or best_distance > max_distance:
                max_distance = best_distance
                best_color = color
        return best_color

    @staticmethod
    def __get_random_color():
        return Color(random.random(), random.random(), random.random(), 1.0)

    @staticmethod
    def __color_distance(c1, c2):
        return sum([abs(x[0] - x[1]) for x in zip(c1.value, c2.value)])

    @property
    def r(self):
        return self.value[0]

    @property
    def g(self):
        return self.value[1]

    @property
    def b(self):
        return self.value[2]

    @property
    def a(self):
        return self.value[3]

    def svg_color(self):
        return 'rgb({:.8f},{:.8f},{:.8f})'.format(self.r * 255.0, self.g * 255.0,
                                                  self.b * 255.0)


COLORS = {
    'transparent': Color(0, 0, 0, 0),
    'black': Color(0.0, 0.0, 0.0, 1.0),
    'white': Color(1.0, 1.0, 1.0, 1.0),
    'transluscent_white': Color(1.0, 1.0, 1.0, 0.7),
    'purple': Color(0.5, 0.0, 0.5, 1.0),
    'light_blue': Color(0.529, 0.808, 0.922, 1.0),
    'blue': Color(0.0, 0.0, 1.0, 1.0),
    'dark_blue': Color(0.0, 0.0, 0.54, 1.0),
    'green': Color(0.0, 0.5, 0.0, 1.0),
    'dark_green': Color(0.0, 0.39, 0.0, 1.0),
    'yellow': Color(1.0, 1.0, 0.0, 1.0),
    'gold': Color(1.0, 0.84, 0.0, 1.0),
    'orange': Color(1.0, 0.647, 0.0, 1.0),
    'red': Color(1.0, 0.0, 0.0, 1.0),
    'dark_red': Color(0.54, 0.0, 0.0, 1.0),
    'brown': Color(0.54, 0.27, 0.1, 1.0),
    'pink': Color(1.0, 0.75, 0.79, 1.0),
    'gray': Color(0.5, 0.5, 0.5, 1.0),
    'dark_gray': Color(0.66, 0.66, 0.66, 1.0),
    'dark_gray_transluscent': Color(0.66, 0.66, 0.66, 0.7843)
}


class Point:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def svg_point_string(self):
        return "{:.8f},{:.8f}".format(self.x, self.y)


class Rect(object):
    origin = Point(0, 0)
    width = 0.0
    height = 0.0

    def __init__(self, origin=Point(0, 0), width=0.0, height=0.0):
        self.origin = origin
        self.width = width
        self.height = height

    def normalized(self):
        normalized_origin = Point(
            self.origin.x + (self.width if self.width < 0 else 0),
            self.origin.y + (self.height if self.height < 0 else 0),
        )
        normalized_width = abs(self.width)
        normalized_height = abs(self.height)
        return Rect(normalized_origin, normalized_width, normalized_height)

    def min_x(self):
        return self.normalized().origin.x

    def min_y(self):
        return self.normalized().origin.y

    def mid_x(self):
        return self.origin.x + (self.width * 0.5)

    def mid_y(self):
        return self.origin.y + (self.height * 0.5)

    def max_x(self):
        norm = self.normalized()
        return norm.origin.x + norm.width

    def max_y(self):
        norm = self.normalized()
        return norm.origin.y + norm.height

    def contract(self, distance):
        self.origin.x += distance
        self.origin.y += distance
        self.width -= 2.0 * distance
        self.height -= 2.0 * distance


def convert_point_to_svg_coordinates(point, image_height):
    y = image_height - point.y
    return Point(point.x, y)


def convert_rect_to_svg_coordinates(rect, image_height):
    """Convert to SVG coordinate system (0,0 at top-left)"""
    normalized_rect = rect.normalized()
    normalized_rect.origin = convert_point_to_svg_coordinates(
        normalized_rect.origin, image_height)
    normalized_rect.height *= -1
    return normalized_rect.normalized()


class SVGWriter:

    def __init__(self, image_width=2406.0, image_height=1054.0, image_margin=20.0,
                 arrow_margin=10.0, arrow_label_margin=5.0, font_size=15.0,
                 font_family='sans-serif'):
        self.image_width = image_width
        self.image_height = image_height
        self.image_margin = image_margin
        self.arrow_margin = arrow_margin
        self.arrow_label_margin = arrow_label_margin
        self.font_size = font_size
        self.text_margin = 0.5 * font_size
        self.font_family = font_family

        self.all_clips_data = []
        self.trackwise_clip_count = []
        self.tracks_duration = []
        self.track_transition_available = []
        self.max_total_duration = 0
        self.global_min_time = 0
        self.global_max_time = 0
        self.scale_x = 1.0
        self.scale_y = 1.0
        self.x_origin = 0
        self.clip_rect_height = 0
        self.vertical_drawing_index = -1
        self.svg_elem = Element("svg",
                                {
                                    "height": "{:.8f}".format(self.image_height),
                                    "width": "{:.8f}".format(self.image_width),
                                    "version": "4.0",
                                    "xmlns": "http://www.w3.org/2000/svg",
                                    "xmlns:xlink": "http://www.w3.org/1999/xlink",
                                })

        # white background
        SubElement(self.svg_elem, "rect",
                   {
                       "width": "100%",
                       "height": "100%",
                       "fill": "white",
                   })

    def draw_rect(self, rect, stroke_width=2.0, stroke_color=COLORS['black']):
        svg_rect = convert_rect_to_svg_coordinates(rect, self.image_height)
        SubElement(self.svg_elem, "rect",
                   {
                       "x": "{:.8f}".format(svg_rect.origin.x),
                       "y": "{:.8f}".format(svg_rect.origin.y),
                       "width": "{:.8f}".format(svg_rect.width),
                       "height": "{:.8f}".format(svg_rect.height),
                       "style": "fill:rgb(255,255,255);stroke-width:{:.8f};"
                                "stroke:{};opacity:1;fill-opacity:0;".format(
                           stroke_width,
                           stroke_color.svg_color())})

    def draw_labeled_rect(self, rect, stroke_width=2.0,
                          stroke_color=COLORS['black'],
                          fill_color=COLORS['white'],
                          label='',
                          label_size=10.0):
        svg_rect = convert_rect_to_svg_coordinates(rect, self.image_height)
        g_elem = SubElement(self.svg_elem, "g",
                            {
                                "transform": "translate({:.8f},{:.8f})".format(
                                    svg_rect.origin.x,
                                    svg_rect.origin.y)
                            })
        SubElement(g_elem, "rect",
                   {
                       "width": "{:.8f}".format(svg_rect.width),
                       "height": "{:.8f}".format(svg_rect.height),
                       "style": "fill:{};stroke-width:{:.8f};"
                                "stroke:{};opacity:1;".format(
                           fill_color.svg_color(),
                           stroke_width,
                           stroke_color.svg_color())
                   })
        sub_svg_elem = SubElement(g_elem, "svg",
                                  {
                                      "width": "{:.8f}".format(svg_rect.width),
                                      "height": "{:.8f}".format(svg_rect.height)
                                  })
        text_elem = SubElement(sub_svg_elem, "text",
                               {
                                   "x": "50%",
                                   "y": "50%",
                                   "font-size": "{:.8f}".format(label_size),
                                   "font-family": self.font_family,
                                   "style":
                                       "stroke:{};stroke-width:{:.8f};"
                                       "fill:{};opacity:{:.8f};".format(
                                           COLORS['black'].svg_color(),
                                           stroke_width / 4.0,
                                           COLORS['black'].svg_color(),
                                           COLORS['black'].a),
                                   "alignment-baseline": "middle",
                                   "text-anchor": "middle"})
        text_elem.text = label

    def draw_dashed_rect(self, rect, stroke_width=2.0, stroke_color=COLORS['black'],
                         fill_color=COLORS['white']):
        svg_rect = convert_rect_to_svg_coordinates(rect, self.image_height)
        SubElement(self.svg_elem, "rect",
                   {
                       "x": "{:.8f}".format(svg_rect.origin.x),
                       "y": "{:.8f}".format(svg_rect.origin.y),
                       "width": "{:.8f}".format(svg_rect.width),
                       "height": "{:.8f}".format(svg_rect.height),
                       "stroke-dasharray": "5",
                       "style": "fill:{};stroke-width:{:.8f};stroke:{};"
                                "opacity:1;fill-opacity:{:.8f}".format(
                           fill_color.svg_color(),
                           stroke_width,
                           stroke_color.svg_color(),
                           fill_color.a)
                   })

    def draw_labeled_dashed_rect_with_border(self, rect, stroke_width=2.0,
                                             fill_color=COLORS['white'],
                                             border_color=COLORS['black'],
                                             label='',
                                             label_size=10.0):
        svg_rect = convert_rect_to_svg_coordinates(rect, self.image_height)
        g_elem = SubElement(self.svg_elem, "g",
                            {
                                "transform": "translate({:.8f},{:.8f})".format(
                                    svg_rect.origin.x, svg_rect.origin.y)
                            })
        SubElement(g_elem, "rect",
                   {
                       "width": "{:.8f}".format(svg_rect.width),
                       "height": "{:.8f}".format(svg_rect.height),
                       "stroke-dasharray": "5",
                       "style": "fill:{};stroke-width:{:.8f};"
                                "stroke:{};opacity:{:.8f};".format(
                           fill_color.svg_color(),
                           stroke_width,
                           border_color.svg_color(),
                           fill_color.a)
                   })
        sub_svg_elem = SubElement(g_elem, "svg",
                                  {
                                      "width": "{:.8f}".format(svg_rect.width),
                                      "height": "{:.8f}".format(svg_rect.height)
                                  })
        text_elem = SubElement(sub_svg_elem, "text",
                               {
                                   "x": "50%",
                                   "y": "50%",
                                   "font-size": "{:.8f}".format(label_size),
                                   "font-family": self.font_family,
                                   "style": "stroke:{};stroke-width:{:.8f};"
                                            "fill:{};opacity:{:.8f};".format(
                                       COLORS['black'].svg_color(),
                                       stroke_width / 4.0,
                                       COLORS['black'].svg_color(),
                                       COLORS['black'].a),
                                   "alignment-baseline": "middle",
                                   "text-anchor": "middle"
                               })
        text_elem.text = label

    def draw_solid_rect(self, rect, fill_color=COLORS['white']):
        svg_rect = convert_rect_to_svg_coordinates(rect, self.image_height)
        SubElement(self.svg_elem, "rect",
                   {
                       "x": "{:.8f}".format(svg_rect.origin.x),
                       "y": "{:.8f}".format(svg_rect.origin.y),
                       "width": "{:.8f}".format(svg_rect.width),
                       "height": "{:.8f}".format(svg_rect.height),
                       "style": "fill:{};stroke-width:0;"
                                "stroke:rgb(0,0,0);opacity:{:.8f};".format(
                           fill_color.svg_color(),
                           fill_color.a)
                   })

    def draw_solid_rect_with_border(self, rect, stroke_width=2.0,
                                    fill_color=COLORS['white'],
                                    border_color=COLORS['black']):
        svg_rect = convert_rect_to_svg_coordinates(rect, self.image_height)
        SubElement(self.svg_elem, "rect",
                   {
                       "x": "{:.8f}".format(svg_rect.origin.x),
                       "y": "{:.8f}".format(svg_rect.origin.y),
                       "width": "{:.8f}".format(svg_rect.width),
                       "height": "{:.8f}".format(svg_rect.height),
                       "style": "fill:{};stroke-width:{:.8f};"
                                "stroke:{};opacity:{:.8f};".format(
                           fill_color.svg_color(),
                           stroke_width,
                           border_color.svg_color(),
                           fill_color.a)
                   })

    def draw_labeled_solid_rect_with_border(self, rect, stroke_width=2.0,
                                            fill_color=COLORS['white'],
                                            border_color=COLORS['black'],
                                            label='',
                                            label_size=10.0):
        svg_rect = convert_rect_to_svg_coordinates(rect, self.image_height)
        g_elem = SubElement(self.svg_elem, "g",
                            {
                                "transform": "translate({:.8f},{:.8f})".format(
                                    svg_rect.origin.x, svg_rect.origin.y)
                            })
        SubElement(g_elem, "rect",
                   {
                       "width": "{:.8f}".format(svg_rect.width),
                       "height": "{:.8f}".format(svg_rect.height),
                       "style": "fill:{};stroke-width:{:.8f};"
                                "stroke:{};opacity:{:.8f};".format(
                           fill_color.svg_color(),
                           stroke_width,
                           border_color.svg_color(),
                           fill_color.a)
                   })
        sub_svg_elem = SubElement(g_elem, "svg",
                                  {
                                      "width": "{:.8f}".format(svg_rect.width),
                                      "height": "{:.8f}".format(svg_rect.height)
                                  })
        text_elem = SubElement(sub_svg_elem, "text",
                               {
                                   "x": "50%",
                                   "y": "50%",
                                   "font-size": "{:.8f}".format(label_size),
                                   "font-family": self.font_family,
                                   "style": "stroke:{};stroke-width:{:.8f};"
                                            "fill:{};opacity:{:.8f};".format(
                                       COLORS['black'].svg_color(),
                                       stroke_width / 4.0,
                                       COLORS['black'].svg_color(),
                                       COLORS['black'].a),
                                   "alignment-baseline": "middle",
                                   "text-anchor": "middle"})
        text_elem.text = label

    def draw_line(self, start_point, end_point, stroke_width,
                  stroke_color=COLORS['black'], is_dashed=False):
        point1 = convert_point_to_svg_coordinates(start_point, self.image_height)
        point2 = convert_point_to_svg_coordinates(end_point, self.image_height)
        style_str = "stroke-width:{:.8f};stroke:{}" \
                    ";opacity:{:.8f};" \
                    "stroke-linecap:butt;".format(stroke_width,
                                                  stroke_color.svg_color(),
                                                  stroke_color.a)
        if is_dashed:
            style_str = style_str + "stroke-dasharray:4 1"
        SubElement(self.svg_elem, "line",
                   {
                       "x1": "{:.8f}".format(point1.x),
                       "y1": "{:.8f}".format(point1.y),
                       "x2": "{:.8f}".format(point2.x),
                       "y2": "{:.8f}".format(point2.y),
                       "style": style_str
                   })

    def draw_arrow(self, start_point, end_point, stroke_width,
                   stroke_color=COLORS['black']):
        point1 = convert_point_to_svg_coordinates(start_point, self.image_height)
        point2 = convert_point_to_svg_coordinates(end_point, self.image_height)
        direction = Point(point2.x - point1.x, point2.y - point1.y)
        direction_magnitude = math.sqrt(direction.x * direction.x +
                                        direction.y * direction.y)
        inv_magnitude = 1.0 / direction_magnitude
        arrowhead_length = 9.0
        arrowhead_half_width = arrowhead_length * 0.5
        direction = Point(direction.x * inv_magnitude, direction.y * inv_magnitude)
        point2 = Point(point2.x - arrowhead_length * direction.x,
                       point2.y - arrowhead_length * direction.y)
        triangle_tip = Point(point2.x + arrowhead_length * direction.x,
                             point2.y + arrowhead_length * direction.y)
        perpendicular_dir = Point(-direction.y, direction.x)
        triangle_pt_1 = Point(point2.x + arrowhead_half_width * perpendicular_dir.x,
                              point2.y + arrowhead_half_width * perpendicular_dir.y)
        triangle_pt_2 = Point(point2.x - arrowhead_half_width * perpendicular_dir.x,
                              point2.y - arrowhead_half_width * perpendicular_dir.y)
        SubElement(self.svg_elem, "line",
                   {
                       "x1": "{:.8f}".format(point1.x),
                       "y1": "{:.8f}".format(point1.y),
                       "x2": "{:.8f}".format(point2.x),
                       "y2": "{:.8f}".format(point2.y),
                       "style": "stroke-width:{:.8f};stroke:{};opacity:{:.8f};"
                                "stroke-linecap:butt;".format(
                           stroke_width,
                           stroke_color.svg_color(),
                           stroke_color.a)
                   })
        SubElement(self.svg_elem, "polygon",
                   {
                       "points": " ".join(p.svg_point_string() for p in
                                          [triangle_tip, triangle_pt_1, triangle_pt_2]),
                       "style": "fill:{};".format(stroke_color.svg_color())
                   })

    def draw_text(self, text, location,
                  text_size, color=COLORS['black'], stroke_width=1.0):
        location_svg = convert_point_to_svg_coordinates(location, self.image_height)
        text_elem = SubElement(self.svg_elem, "text",
                               {
                                   "x": "{:.8f}".format(location_svg.x),
                                   "y": "{:.8f}".format(location_svg.y),
                                   "font-size": "{:.8f}".format(text_size),
                                   "font-family": self.font_family,
                                   "style": "stroke:{};stroke-width:{:.8f};"
                                            "fill:{};opacity:{:.8f};".format(
                                       color.svg_color(),
                                       stroke_width / 4.0,
                                       color.svg_color(),
                                       color.a)
                               })
        text_elem.text = text

    def get_image(self):
        # Python 3 produces a bytestring with the tostring() method, whereas Python 2
        # gives an str object. The try-except block below checks for this case.
        xmlstr = tostring(self.svg_elem, encoding='utf-8', method='xml')
        try:
            xmlstr = xmlstr.decode("utf8")
        except UnicodeDecodeError:
            pass
        return minidom.parseString(xmlstr).toprettyxml(indent='  ')


class ClipData(object):

    def __init__(self, src_start=0.0, src_end=0.0, avlbl_start=0.0,
                 avlbl_end=0.0, avlbl_duration=0.0,
                 trim_start=0.0, trim_duration=0.0, clip_id=0,
                 transition_begin=None, transition_end=None):
        self.src_start = src_start
        self.src_end = src_end
        self.avlbl_start = avlbl_start
        self.avlbl_end = avlbl_end
        self.avlbl_duration = avlbl_duration
        self.trim_start = trim_start
        self.trim_duration = trim_duration
        self.clip_id = clip_id
        self.transition_begin = transition_begin
        self.transition_end = transition_end


def draw_item(otio_obj, svg_writer, extra_data=()):
    WRITE_TYPE_MAP = {
        otio.schema.Timeline: _draw_timeline,
        otio.schema.Stack: _draw_stack,
        otio.schema.Track: _draw_track,
        otio.schema.Clip: _draw_clip,
        otio.schema.Gap: _draw_gap,
        otio.schema.Transition: _draw_transition,
        otio.schema.SerializableCollection: _draw_collection,
    }
    if type(otio_obj) in WRITE_TYPE_MAP:
        return WRITE_TYPE_MAP[type(otio_obj)](otio_obj, svg_writer, extra_data)


# Draw Timeline and calculate Clip and Gap data
def _draw_timeline(timeline, svg_writer, extra_data=()):
    clip_count = 0
    transition_track_count = 0
    for track in timeline.tracks:
        if len(track) == 0:
            continue
        current_track_clips_data = []
        current_track_has_transition = False
        current_transition = None
        track_duration = 0
        min_time = 0
        max_time = 0
        for item in track:
            if isinstance(item, otio.schema.Transition):
                current_track_has_transition = True
                current_transition = item
                current_track_clips_data[-1].transition_end = item
                continue
            avlbl_start = track_duration - item.trimmed_range().start_time.value
            if isinstance(item, otio.schema.Clip):
                avlbl_start += item.available_range().start_time.value
            min_time = min(min_time, avlbl_start)
            src_start = track_duration
            track_duration += item.trimmed_range().duration.value
            src_end = track_duration - 1
            avlbl_end = 0.0
            trim_start = item.trimmed_range().start_time.value
            trim_duration = item.trimmed_range().duration.value
            if isinstance(item, otio.schema.Clip):
                avlbl_end = (item.available_range().start_time.value +
                             item.available_range().duration.value -
                             item.trimmed_range().start_time.value -
                             item.trimmed_range().duration.value + track_duration - 1)
                clip_count += 1
                avlbl_duration = item.available_range().duration.value
                clip_data = ClipData(src_start, src_end, avlbl_start,
                                     avlbl_end, avlbl_duration, trim_start,
                                     trim_duration,
                                     clip_count - 1)
                if current_transition is not None:
                    clip_data.transition_begin = current_transition
                    current_transition = None
                current_track_clips_data.append(clip_data)
            elif isinstance(item, otio.schema.Gap):
                avlbl_end = src_end
                avlbl_duration = trim_duration
                current_transition = None
                clip_data = ClipData(src_start, src_end, avlbl_start,
                                     avlbl_end, avlbl_duration, trim_start,
                                     trim_duration,
                                     "Gap", -1)
                current_track_clips_data.append(clip_data)
            max_time = max(max_time, avlbl_end)
        svg_writer.global_max_time = max(svg_writer.global_max_time, max_time)
        svg_writer.global_min_time = min(svg_writer.global_min_time, min_time)
        svg_writer.all_clips_data.append(current_track_clips_data)
        svg_writer.tracks_duration.append(track_duration)
        svg_writer.track_transition_available.append(current_track_has_transition)
        if current_track_has_transition:
            transition_track_count += 1
        # store track-wise clip count to draw arrows from stack to tracks
        if len(svg_writer.trackwise_clip_count) == 0:
            svg_writer.trackwise_clip_count.append(clip_count)
        else:
            svg_writer.trackwise_clip_count.append(
                clip_count - svg_writer.trackwise_clip_count[
                    len(svg_writer.trackwise_clip_count) - 1])
    # The scale in x direction is calculated considering margins on the
    # left and right side if the image
    svg_writer.scale_x = (svg_writer.image_width - (2.0 * svg_writer.image_margin)) / \
                         (svg_writer.global_max_time - svg_writer.global_min_time + 1.0)
    svg_writer.x_origin = ((-svg_writer.global_min_time) * svg_writer.scale_x +
                           svg_writer.image_margin)
    track_count = len(svg_writer.tracks_duration)
    # The rect height is calculated considering the following:
    # Total space available:
    #   image height - top & bottom margin -
    #   space for two labels for the bottom-most rect
    # Number of total rects to fit the height of the drawing space:
    #   track_count * 2.0 = one for track rect and one for the sequence of
    #                       components on that track
    #   + 2.0 = timeline and stack rects
    #   clip_count = we need to draw a rect for a media reference per clip
    #   transition_track_count = we need one more row per the number of tracks with
    #                            transitions
    # NumberOfRects * 2.0 - 1.0 = to account for "one rect space" between all the rects
    total_image_margin_space = 2.0 * svg_writer.image_margin
    bottom_label_space = 2.0 * svg_writer.font_size
    svg_total_draw_space = (svg_writer.image_height - total_image_margin_space -
                            bottom_label_space)
    track_sequence_rect_count = track_count * 2.0
    timeline_stack_rect_count = 2.0
    rect_count = (track_sequence_rect_count + timeline_stack_rect_count +
                  clip_count + transition_track_count)
    total_slots = rect_count * 2.0 - 1.0
    svg_writer.clip_rect_height = svg_total_draw_space / total_slots

    # Draw Timeline
    svg_writer.vertical_drawing_index += 2
    timeline_origin = Point(svg_writer.x_origin,
                            svg_writer.image_height - svg_writer.image_margin -
                            svg_writer.vertical_drawing_index *
                            svg_writer.clip_rect_height)
    svg_writer.max_total_duration = max(svg_writer.tracks_duration)
    label_text_size = 0.4 * svg_writer.clip_rect_height
    svg_writer.draw_labeled_solid_rect_with_border(
        Rect(timeline_origin, svg_writer.max_total_duration * svg_writer.scale_x,
             svg_writer.clip_rect_height), label="Timeline", label_size=label_text_size)
    time_marker_height = 0.15 * svg_writer.clip_rect_height
    for i in range(1, int(svg_writer.max_total_duration)):
        start_pt = Point(svg_writer.x_origin + (i * svg_writer.scale_x),
                         timeline_origin.y)
        end_pt = Point(start_pt.x, start_pt.y + time_marker_height)
        svg_writer.draw_line(start_point=start_pt, end_point=end_pt, stroke_width=1.0,
                             stroke_color=COLORS['black'])
    # Draw arrow from timeline to stack
    timeline_width = svg_writer.max_total_duration * svg_writer.scale_x
    arrow_start = Point(svg_writer.x_origin + timeline_width * 0.5,
                        timeline_origin.y - svg_writer.arrow_margin)
    arrow_end = Point(svg_writer.x_origin + timeline_width * 0.5,
                      timeline_origin.y - svg_writer.clip_rect_height +
                      svg_writer.arrow_margin)
    svg_writer.draw_arrow(start_point=arrow_start, end_point=arrow_end,
                          stroke_width=2.0, stroke_color=COLORS['black'])
    arrow_label_location = Point(arrow_start.x + svg_writer.arrow_label_margin,
                                 (arrow_start.y + arrow_end.y) * 0.5)
    svg_writer.draw_text('tracks', arrow_label_location, svg_writer.font_size)
    # Draw global_start_time info
    if timeline.global_start_time is None:
        start_time_text = r'global_start_time: {}'.format('None')
    else:
        start_time_text = r'global_start_time: {}'.format(
            repr(float(round(timeline.global_start_time.value, 1))))
    start_time_location = Point(timeline_origin.x + svg_writer.font_size,
                                timeline_origin.y - svg_writer.font_size)
    svg_writer.draw_text(start_time_text, start_time_location, svg_writer.font_size)

    # Draw stack
    draw_item(timeline.tracks, svg_writer,
              (svg_writer.x_origin, svg_writer.max_total_duration))


# Draw stack
def _draw_stack(stack, svg_writer, extra_data=()):
    stack_x_origin = extra_data[0]
    stack_duration = extra_data[1]
    svg_writer.vertical_drawing_index += 2
    stack_origin = Point(stack_x_origin,
                         svg_writer.image_height - svg_writer.image_margin -
                         svg_writer.vertical_drawing_index *
                         svg_writer.clip_rect_height)
    stack_text_size = 0.4 * svg_writer.clip_rect_height
    svg_writer.draw_labeled_solid_rect_with_border(
        Rect(stack_origin, stack_duration * svg_writer.scale_x,
             svg_writer.clip_rect_height),
        label="Stack", fill_color=COLORS['dark_gray_transluscent'],
        label_size=stack_text_size)
    time_marker_height = 0.15 * svg_writer.clip_rect_height
    for i in range(1, int(svg_writer.max_total_duration)):
        start_pt = Point(svg_writer.x_origin + (i * svg_writer.scale_x), stack_origin.y)
        end_pt = Point(start_pt.x, start_pt.y + time_marker_height)
        svg_writer.draw_line(start_point=start_pt, end_point=end_pt, stroke_width=1.0,
                             stroke_color=COLORS['black'])
    for i in range(0, len(svg_writer.tracks_duration)):
        draw_item(stack[i], svg_writer, (stack_x_origin, svg_writer.tracks_duration[i],
                                         svg_writer.all_clips_data[i],
                                         svg_writer.track_transition_available[i]))
    # Draw arrows from stack to tracks
    #   arrow from stack to first track
    stack_width = stack_duration * svg_writer.scale_x
    arrow_start = Point(svg_writer.x_origin + stack_width * 0.5,
                        stack_origin.y - svg_writer.arrow_margin)
    arrow_end = Point(svg_writer.x_origin + stack_width * 0.5,
                      stack_origin.y - svg_writer.clip_rect_height +
                      svg_writer.arrow_margin)
    svg_writer.draw_arrow(start_point=arrow_start, end_point=arrow_end,
                          stroke_width=2.0,
                          stroke_color=COLORS['black'])
    end_arrow_offset = 1
    #   arrows from stack to rest of the tracks
    for i in range(1, len(svg_writer.trackwise_clip_count)):
        arrow_x_increment_per_track = 10.0
        end_arrow_offset += (svg_writer.trackwise_clip_count[i - 1] * 2.0 + 4.0)
        arrow_start = Point(
            (i * arrow_x_increment_per_track) + svg_writer.x_origin + stack_width * 0.5,
            stack_origin.y - svg_writer.arrow_margin)
        arrow_end = Point(
            (i * arrow_x_increment_per_track) + svg_writer.x_origin + stack_width * 0.5,
            stack_origin.y - (end_arrow_offset * svg_writer.clip_rect_height) +
            svg_writer.arrow_margin)
        svg_writer.draw_arrow(start_point=arrow_start, end_point=arrow_end,
                              stroke_width=2.0,
                              stroke_color=COLORS['black'])
    arrow_label_text = r'children[{}]'.format(len(svg_writer.trackwise_clip_count))
    arrow_label_location = Point(arrow_start.x + svg_writer.arrow_label_margin,
                                 stack_origin.y - svg_writer.clip_rect_height * 0.5)
    svg_writer.draw_text(arrow_label_text, arrow_label_location, svg_writer.font_size)
    # Draw range info
    if stack.trimmed_range() is None:
        trimmed_range_text = r'trimmed_range() -> {}'.format('None')
    else:
        trimmed_range_text = r'trimmed_range() -> {}, {}'.format(
            repr(float(round(stack.trimmed_range().start_time.value, 1))),
            repr(float(round(stack.trimmed_range().duration.value, 1))))
    if stack.source_range is None:
        source_range_text = r'source_range: {}'.format('None')
    else:
        source_range_text = r'source_range: {}, {}'.format(
            repr(float(round(stack.source_range.start_time.value, 1))),
            repr(float(round(stack.source_range.duration.value, 1))))
    trimmed_range_location = Point(stack_origin.x + svg_writer.font_size,
                                   stack_origin.y + svg_writer.clip_rect_height +
                                   svg_writer.text_margin)
    source_range_location = Point(stack_origin.x + svg_writer.font_size,
                                  stack_origin.y - svg_writer.font_size)
    svg_writer.draw_text(trimmed_range_text, trimmed_range_location,
                         svg_writer.font_size,
                         )
    svg_writer.draw_text(source_range_text, source_range_location, svg_writer.font_size)


def _draw_track(track, svg_writer, extra_data=()):
    svg_writer.vertical_drawing_index += 2
    track_x_origin = extra_data[0]
    track_duration = extra_data[1]
    clips_data = extra_data[2]
    track_has_transition = extra_data[3]
    track_origin = Point(track_x_origin,
                         svg_writer.image_height - svg_writer.image_margin -
                         svg_writer.vertical_drawing_index *
                         svg_writer.clip_rect_height)
    track_text_size = 0.4 * svg_writer.clip_rect_height
    track_text = track.name if track.name else 'Track'
    svg_writer.draw_labeled_solid_rect_with_border(
        Rect(track_origin, track_duration * svg_writer.scale_x,
             svg_writer.clip_rect_height),
        label=track_text, fill_color=COLORS['dark_gray_transluscent'],
        label_size=track_text_size)
    time_marker_height = 0.15 * svg_writer.clip_rect_height
    for i in range(1, int(track_duration)):
        start_pt = Point(svg_writer.x_origin + (i * svg_writer.scale_x), track_origin.y)
        end_pt = Point(start_pt.x, start_pt.y + time_marker_height)
        svg_writer.draw_line(start_point=start_pt, end_point=end_pt, stroke_width=1.0,
                             stroke_color=COLORS['black'])
    item_count = 0
    clip_count = 0
    transition_count = 0
    svg_writer.vertical_drawing_index += 2
    if track_has_transition:
        svg_writer.vertical_drawing_index += 2
    for item in track:
        if isinstance(item, otio.schema.Clip):
            clip_count += 1
            draw_item(item, svg_writer, (clips_data[item_count], clip_count))
            item_count += 1
        elif isinstance(item, otio.schema.Gap):
            draw_item(item, svg_writer, (clips_data[item_count],))
            item_count += 1
        elif isinstance(item, otio.schema.Transition):
            cut_x = svg_writer.x_origin + (clips_data[clip_count].src_start *
                                           svg_writer.scale_x)
            draw_item(item, svg_writer, (cut_x,))
            transition_count += 1
    svg_writer.vertical_drawing_index += (2 * clip_count)
    # Draw arrow from track to clips
    track_width = track_duration * svg_writer.scale_x
    arrow_start = Point(svg_writer.x_origin + track_width * 0.5,
                        track_origin.y - svg_writer.arrow_margin)
    arrow_end = Point(svg_writer.x_origin + track_width * 0.5,
                      track_origin.y - svg_writer.clip_rect_height +
                      svg_writer.arrow_margin)
    svg_writer.draw_arrow(start_point=arrow_start, end_point=arrow_end,
                          stroke_width=2.0,
                          stroke_color=COLORS['black'])
    arrow_label_text = r'children[{}]'.format(item_count + transition_count)
    arrow_label_location = Point(arrow_start.x + svg_writer.arrow_label_margin,
                                 track_origin.y - svg_writer.clip_rect_height * 0.5)
    svg_writer.draw_text(arrow_label_text, arrow_label_location, svg_writer.font_size)
    # Draw range info
    if track.trimmed_range() is None:
        trimmed_range_text = r'trimmed_range() -> {}'.format('None')
    else:
        trimmed_range_text = r'trimmed_range() -> {}, {}'.format(
            repr(float(round(track.trimmed_range().start_time.value, 1))),
            repr(float(round(track.trimmed_range().duration.value, 1))))
    if track.source_range is None:
        source_range_text = r'source_range: {}'.format('None')
    else:
        source_range_text = r'source_range: {}, {}'.format(
            repr(float(round(track.source_range.start_time.value, 1))),
            repr(float(round(track.source_range.duration.value, 1))))
    trimmed_range_location = Point(track_origin.x + svg_writer.font_size,
                                   track_origin.y + svg_writer.clip_rect_height +
                                   svg_writer.text_margin)
    source_range_location = Point(track_origin.x + svg_writer.font_size,
                                  track_origin.y - svg_writer.font_size)
    svg_writer.draw_text(trimmed_range_text, trimmed_range_location,
                         svg_writer.font_size,
                         )
    svg_writer.draw_text(source_range_text, source_range_location, svg_writer.font_size)


# Draw clip
def _draw_clip(clip, svg_writer, extra_data=()):
    clip_data = extra_data[0]
    clip_count = extra_data[1]
    clip_color = Color.random_color()
    clip_origin = Point(
        svg_writer.x_origin + (clip_data.src_start * svg_writer.scale_x),
        svg_writer.image_height - svg_writer.image_margin -
        svg_writer.vertical_drawing_index * svg_writer.clip_rect_height)
    clip_rect = Rect(clip_origin, clip_data.trim_duration * svg_writer.scale_x,
                     svg_writer.clip_rect_height)
    clip_text_size = 0.4 * svg_writer.clip_rect_height
    clip_text = r'Clip-{}'.format(clip_data.clip_id) if len(
        clip.name) == 0 else clip.name
    svg_writer.draw_labeled_solid_rect_with_border(
        clip_rect,
        label=clip_text, fill_color=clip_color,
        label_size=clip_text_size)
    time_marker_height = 0.15 * svg_writer.clip_rect_height
    for i in range(int(clip_data.src_start), int(clip_data.src_end) + 1):
        start_pt = Point(svg_writer.x_origin + (i * svg_writer.scale_x), clip_origin.y)
        end_pt = Point(start_pt.x, start_pt.y + time_marker_height)
        svg_writer.draw_line(start_point=start_pt, end_point=end_pt, stroke_width=1.0,
                             stroke_color=COLORS['black'])
    # Draw range info
    if clip.trimmed_range() is None:
        trimmed_range_text = r'trimmed_range() -> {}'.format('None')
    else:
        trimmed_range_text = r'trimmed_range() -> {}, {}'.format(
            repr(float(round(clip.trimmed_range().start_time.value, 1))),
            repr(float(round(clip.trimmed_range().duration.value, 1))))
    if clip.source_range is None:
        source_range_text = r'source_range: {}'.format('None')
    else:
        source_range_text = r'source_range: {}, {}'.format(
            repr(float(round(clip.source_range.start_time.value, 1))),
            repr(float(round(clip.source_range.duration.value, 1))))
    trimmed_range_location = Point(clip_origin.x + svg_writer.font_size,
                                   clip_origin.y + svg_writer.clip_rect_height +
                                   svg_writer.text_margin)
    source_range_location = Point(clip_origin.x + svg_writer.font_size,
                                  clip_origin.y - svg_writer.font_size)
    svg_writer.draw_text(trimmed_range_text, trimmed_range_location,
                         svg_writer.font_size,
                         )
    svg_writer.draw_text(source_range_text, source_range_location, svg_writer.font_size)

    # Draw media reference
    trim_media_origin = Point(
        svg_writer.x_origin + (clip_data.src_start * svg_writer.scale_x),
        svg_writer.image_height - svg_writer.image_margin -
        (svg_writer.vertical_drawing_index + clip_count * 2) *
        svg_writer.clip_rect_height)
    media_origin = Point(
        svg_writer.x_origin + (clip_data.avlbl_start * svg_writer.scale_x),
        svg_writer.image_height - svg_writer.image_margin -
        (svg_writer.vertical_drawing_index + clip_count * 2) *
        svg_writer.clip_rect_height)
    svg_writer.draw_rect(
        Rect(media_origin, clip_data.avlbl_duration * svg_writer.scale_x,
             svg_writer.clip_rect_height))
    media_text_size = 0.4 * svg_writer.clip_rect_height
    media_text = r'Media-{}'.format(clip_data.clip_id) if len(
        clip.media_reference.name) == 0 else clip.media_reference.name
    svg_writer.draw_labeled_solid_rect_with_border(
        Rect(trim_media_origin, clip_data.trim_duration * svg_writer.scale_x,
             svg_writer.clip_rect_height),
        label=media_text, fill_color=clip_color,
        label_size=media_text_size)
    for i in range(int(clip_data.avlbl_start),
                   int(clip_data.avlbl_end) + 1):
        start_pt = Point(svg_writer.x_origin + (i * svg_writer.scale_x), media_origin.y)
        if start_pt.x < media_origin.x:
            continue
        end_pt = Point(start_pt.x, start_pt.y + time_marker_height)
        svg_writer.draw_line(start_point=start_pt, end_point=end_pt, stroke_width=1.0,
                             stroke_color=COLORS['black'])
    # Draw media_reference info
    if clip.available_range() is None:
        available_range_text = r'available_range: {}'.format('None')
    else:
        available_range_text = r'available_range: {}, {}'.format(
            repr(float(round(clip.available_range().start_time.value, 1))),
            repr(float(round(clip.available_range().duration.value, 1))))
    available_range_location = Point(media_origin.x + svg_writer.font_size,
                                     media_origin.y - svg_writer.font_size)
    svg_writer.draw_text(available_range_text, available_range_location,
                         svg_writer.font_size,
                         )
    if hasattr(clip.media_reference, 'target_url'):
        if clip.media_reference.target_url is None:
            target_url_text = r'target_url: {}'.format('Media Unavailable')
        else:
            target_url_text = r'target_url: {}'.format(clip.media_reference.target_url)
        target_url_location = Point(media_origin.x + svg_writer.font_size,
                                    media_origin.y - 2.0 * svg_writer.font_size)
        svg_writer.draw_text(target_url_text, target_url_location, svg_writer.font_size)
    # Draw arrow from clip to media reference
    clip_media_height_difference = (((clip_count - 1) * 2.0 + 1) *
                                    svg_writer.clip_rect_height)
    media_arrow_start = Point(
        clip_origin.x + (clip_data.trim_duration * svg_writer.scale_x) * 0.5,
        clip_origin.y - svg_writer.arrow_margin)
    media_arrow_end = Point(
        clip_origin.x + (clip_data.trim_duration * svg_writer.scale_x) * 0.5,
        clip_origin.y - clip_media_height_difference + svg_writer.arrow_margin)
    svg_writer.draw_arrow(start_point=media_arrow_start, end_point=media_arrow_end,
                          stroke_width=2.0, stroke_color=COLORS['black'])
    arrow_label_text = r'media_reference'
    arrow_label_location = Point(media_arrow_start.x + svg_writer.arrow_label_margin,
                                 media_arrow_start.y -
                                 svg_writer.clip_rect_height * 0.5)
    svg_writer.draw_text(arrow_label_text, arrow_label_location, svg_writer.font_size)
    # Draw media transition sections
    if clip_data.transition_end is not None:
        cut_x = clip_origin.x + clip_rect.width
        section_start_pt = Point(cut_x, media_origin.y)
        # Handle the case of transition ending at cut
        if clip_data.transition_end.out_offset.value == 0.0:
            media_transition_rect = Rect(section_start_pt,
                                         -clip_data.transition_end.in_offset.value *
                                         svg_writer.scale_x,
                                         svg_writer.clip_rect_height)
            marker_x = [clip_data.src_end,
                        clip_data.src_end - clip_data.transition_end.in_offset.value]
        else:
            media_transition_rect = Rect(section_start_pt,
                                         clip_data.transition_end.out_offset.value *
                                         svg_writer.scale_x,
                                         svg_writer.clip_rect_height)
            marker_x = [clip_data.src_end,
                        clip_data.src_end + clip_data.transition_end.out_offset.value]
        section_color = Color(clip_color[0], clip_color[1], clip_color[2], 0.5)
        svg_writer.draw_dashed_rect(media_transition_rect, fill_color=section_color)
        marker_x.sort()
        # Draw markers for transition sections
        for i in range(int(marker_x[0]),
                       int(marker_x[1]) + 1):
            start_pt = Point(svg_writer.x_origin + (i * svg_writer.scale_x),
                             media_origin.y)
            if start_pt.x < media_transition_rect.min_x():
                continue
            end_pt = Point(start_pt.x, start_pt.y + time_marker_height)
            svg_writer.draw_line(start_point=start_pt, end_point=end_pt,
                                 stroke_width=1.0,
                                 stroke_color=COLORS['black'])
    if clip_data.transition_begin is not None:
        cut_x = clip_origin.x
        section_start_pt = Point(cut_x, media_origin.y)
        # Handle the case of transition starting at cut
        if clip_data.transition_begin.in_offset.value == 0.0:
            media_transition_rect = Rect(section_start_pt,
                                         clip_data.transition_begin.out_offset.value *
                                         svg_writer.scale_x,
                                         svg_writer.clip_rect_height)
            marker_x = [clip_data.src_start,
                        clip_data.src_start +
                        clip_data.transition_begin.out_offset.value]
        else:
            media_transition_rect = Rect(section_start_pt,
                                         -clip_data.transition_begin.in_offset.value *
                                         svg_writer.scale_x,
                                         svg_writer.clip_rect_height)
            marker_x = [clip_data.src_start,
                        clip_data.src_start -
                        clip_data.transition_begin.out_offset.value]
        section_color = Color(clip_color[0], clip_color[1], clip_color[2], 0.5)
        svg_writer.draw_dashed_rect(media_transition_rect, fill_color=section_color)
        marker_x.sort()
        # Draw markers for transition sections
        for i in range(int(marker_x[0]),
                       int(marker_x[1]) + 1):
            start_pt = Point(svg_writer.x_origin + (i * svg_writer.scale_x),
                             media_origin.y)
            if start_pt.x < media_transition_rect.min_x():
                continue
            end_pt = Point(start_pt.x, start_pt.y + 0.15 * svg_writer.clip_rect_height)
            svg_writer.draw_line(start_point=start_pt, end_point=end_pt,
                                 stroke_width=1.0,
                                 stroke_color=COLORS['black'])


def _draw_gap(gap, svg_writer, extra_data=()):
    gap_data = extra_data[0]
    gap_origin = Point(svg_writer.x_origin + (gap_data.src_start * svg_writer.scale_x),
                       svg_writer.image_height - svg_writer.image_margin -
                       svg_writer.vertical_drawing_index * svg_writer.clip_rect_height)
    gap_text_size = 0.4 * svg_writer.clip_rect_height
    gap_text = 'Gap'
    svg_writer.draw_labeled_dashed_rect_with_border(
        Rect(gap_origin, gap_data.trim_duration * svg_writer.scale_x,
             svg_writer.clip_rect_height),
        label=gap_text, label_size=gap_text_size)
    time_marker_height = 0.15 * svg_writer.clip_rect_height
    for i in range(int(gap_data.src_start), int(gap_data.src_end) + 1):
        start_pt = Point(svg_writer.x_origin + (i * svg_writer.scale_x), gap_origin.y)
        end_pt = Point(start_pt.x, start_pt.y + time_marker_height)
        svg_writer.draw_line(start_point=start_pt, end_point=end_pt, stroke_width=1.0,
                             stroke_color=COLORS['black'])
    # Draw range info
    if gap.trimmed_range() is None:
        trimmed_range_text = r'trimmed_range() -> {}'.format('None')
    else:
        trimmed_range_text = r'trimmed_range() -> {}, {}'.format(
            repr(float(round(gap.trimmed_range().start_time.value, 1))),
            repr(float(round(gap.trimmed_range().duration.value, 1))))
    if gap.source_range is None:
        source_range_text = r'source_range: {}'.format('None')
    else:
        source_range_text = r'source_range: {}, {}'.format(
            repr(float(round(gap.source_range.start_time.value, 1))),
            repr(float(round(gap.source_range.duration.value, 1))))
    trimmed_range_location = Point(gap_origin.x + svg_writer.font_size,
                                   gap_origin.y + svg_writer.clip_rect_height +
                                   svg_writer.text_margin)
    source_range_location = Point(gap_origin.x + svg_writer.font_size,
                                  gap_origin.y - svg_writer.font_size)
    svg_writer.draw_text(trimmed_range_text, trimmed_range_location,
                         svg_writer.font_size,
                         )
    svg_writer.draw_text(source_range_text, source_range_location, svg_writer.font_size)


def _draw_transition(transition, svg_writer, extra_data=()):
    cut_x = extra_data[0]
    transition_origin = Point(cut_x - (transition.in_offset.value * svg_writer.scale_x),
                              svg_writer.image_height - svg_writer.image_margin -
                              (svg_writer.vertical_drawing_index - 2) *
                              svg_writer.clip_rect_height)
    transition_rect = Rect(transition_origin,
                           (transition.in_offset.value + transition.out_offset.value) *
                           svg_writer.scale_x,
                           svg_writer.clip_rect_height)
    transition_name = 'Transition' if len(
        transition.name) == 0 else transition.name
    transition_name_size = 0.4 * svg_writer.clip_rect_height
    svg_writer.draw_labeled_rect(transition_rect, label=transition_name,
                                 label_size=transition_name_size)
    line_end = Point(transition_origin.x + transition_rect.width,
                     transition_origin.y + transition_rect.height)
    svg_writer.draw_line(transition_origin, line_end, stroke_width=1.0,
                         stroke_color=COLORS['black'])
    in_offset_location = Point(transition_origin.x + svg_writer.font_size,
                               transition_origin.y - svg_writer.font_size)
    out_offset_location = Point(transition_origin.x + svg_writer.font_size,
                                transition_origin.y - 2.0 * svg_writer.font_size)
    in_offset_text = r'in_offset: ' \
                     r'{}'.format(repr(float(round(transition.in_offset.value, 1))))
    out_offset_text = r'out_offset: ' \
                      r'{}'.format(repr(float(round(transition.out_offset.value, 1))))
    svg_writer.draw_text(in_offset_text, in_offset_location, svg_writer.font_size)
    svg_writer.draw_text(out_offset_text, out_offset_location, svg_writer.font_size)
    cut_location = Point(cut_x, transition_origin.y)
    cut_line_end = Point(cut_x,
                         svg_writer.image_height - svg_writer.image_margin -
                         svg_writer.vertical_drawing_index *
                         svg_writer.clip_rect_height)
    svg_writer.draw_line(cut_location, cut_line_end, stroke_width=1.0,
                         stroke_color=COLORS['black'])


def _draw_collection(collection, svg_writer, extra_data=()):
    pass


def convert_otio_to_svg(timeline, width, height):
    global random_colors_used

    svg_writer = SVGWriter(image_width=width, image_height=height,
                           font_family='sans-serif', image_margin=20.0, font_size=15.0,
                           arrow_label_margin=5.0)
    random_colors_used = []
    random.seed(100)
    draw_item(timeline, svg_writer, ())

    return svg_writer.get_image()


# --------------------
# adapter requirements
# --------------------

def write_to_string(input_otio, width=2406.0, height=1054.0):
    return convert_otio_to_svg(input_otio, width=width, height=height)



#!/usr/bin/env python
#
# SPDX-License-Identifier: Apache-2.0
# Copyright Contributors to the OpenTimelineIO project

"""Unsupported contrib code for OpenTimelineIO."""

# flake8: noqa

from . import (
    adapters
)

__version__ = "0.15.0"
__author__ = "Contributors to the OpenTimelineIO project"
__author_email__ = "otio-discussion@lists.aswf.io"
__license__ = "Modified Apache 2.0 License"
